/*
 * Copyright © 2016 - 2017 Copyright (c) Sprint, Inc. and others.  All rights reserved.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v1.0 which accompanies this distribution,
 * and is available at http://www.eclipse.org/legal/epl-v10.html
 */
package org.onosproject.fpcagent.util.eventStream;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.json.JSONException;
import org.json.JSONObject;
import org.onosproject.restconf.api.RestconfRpcOutput;
import org.onosproject.restconf.api.RestconfService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.URI;
import java.util.AbstractMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * Implements a Worker to process event-data pairs sent by the FPC Client
 */
public class NBEventWorkerManager implements AutoCloseable {
    private static final Logger log = LoggerFactory.getLogger(NBEventWorkerManager.class);
    private static NBEventWorkerManager _instance = null;

    public static Map<Integer, String> clientIdToUri = new ConcurrentHashMap<Integer, String>();

    private final BlockingQueue<Entry<String, Entry<String, String>>> blockingQueue;
    private final int poolSize;
    private boolean run;
    private final RestconfService restconfService;

    protected NBEventWorkerManager(int poolSize, RestconfService restconfService) {
        this.poolSize = poolSize;
        this.restconfService = restconfService;
        this.blockingQueue = new LinkedBlockingQueue<>();
    }

    public static NBEventWorkerManager createInstance(int poolSize, RestconfService restconfService) {
        if (_instance == null) {
            _instance = new NBEventWorkerManager(poolSize, restconfService);
        }
        return _instance;
    }

    public static NBEventWorkerManager getInstance() {
        return _instance;
    }

    public void send(Entry<String, Entry<String, String>> buf) {
        try {
            blockingQueue.put(buf);
        } catch (InterruptedException e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }
    }

    public void open() {
        this.run = true;
        Entry<String, Entry<String, String>> contents = null;
        while (run) {
            try {
                contents = blockingQueue.take();

                // call RESTCONF service
                ObjectMapper mapper = new ObjectMapper();
                ObjectNode node = (ObjectNode) mapper.readTree(contents.getValue().getValue());
                if (contents.getValue().getKey().contains("ietf-dmm-fpcagent:configure")) {
                    CompletableFuture<RestconfRpcOutput> output = restconfService.runRpc(
                            new URI("/onos/restconf/operations/ietf-dmm-fpcagent:configure"),
                            node,
                            contents.getKey()
                    );
                    ConfigureService.blockingQueue.add(
                            new AbstractMap.SimpleEntry(
                                    contents.getKey(),
                                    "event:application/json;/restconf/operations/ietf-dmm-fpcagent:configure\ndata:" + output.get().output().toString() + "\r\n"
                            )
                    );
                } else if (contents.getValue().getKey().contains("fpc:register-client")) {
                    CompletableFuture<RestconfRpcOutput> output = restconfService.runRpc(
                            new URI("/onos/restconf/operations/fpc:register-client"),
                            node,
                            contents.getKey()
                    );
                    addClientIdToUriMapping(output.get().output().toString(), contents.getKey());
                    ConfigureService.registerClientQueue.add(
                            new AbstractMap.SimpleEntry(
                                    contents.getKey(),
                                    "event:application/json;/restconf/operations/fpc:register_client\ndata:" + output.get().output().toString() + "\r\n"
                            )
                    );
                } else if (contents.getValue().getKey().contains("fpc:deregister-client")) {
                    CompletableFuture<RestconfRpcOutput> output = restconfService.runRpc(
                            new URI("/onos/restconf/operations/fpc:deregister-client"),
                            node,
                            contents.getKey()
                    );
                    removeClientIdToUriMapping(contents.getValue().getValue());
                }

            } catch (Exception e) {
                log.error(ExceptionUtils.getFullStackTrace(e));
            }
        }
    }

    /**
     * Add a FPC client URI to FPC client id mapping
     *
     * @param registerClientOutput - Output of the register_cleint rpc
     * @param uri                  - FPC Client URI
     */
    private void addClientIdToUriMapping(String registerClientOutput, String uri) {
        try {
            JSONObject registerJSON = new JSONObject(registerClientOutput);
            clientIdToUri.put(Integer.parseInt(registerJSON.getJSONObject("output").getString("client-id")), uri);
        } catch (JSONException e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }
    }

    /**
     * Remove a FPC client URI to FPCClient Id mapping
     *
     * @param deregisterClientInput - Input String of deregister_client rpc
     */
    private void removeClientIdToUriMapping(String deregisterClientInput) {
        try {
            JSONObject deregisterJSON = new JSONObject(deregisterClientInput);
            clientIdToUri.remove(Integer.parseInt(deregisterJSON.getJSONObject("input").getString("client-id")));
        } catch (JSONException e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }
    }

    @Override
    public void close() {
        this.run = false;
    }
}
